/*
 * 2003  Benedikt Meurer <bmeurer@unix-ag.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>

#include <assert.h>
#include <string.h>

#include "array.h"


/**
 * Check whether item is unique in a[0...(pos-1)]
 */
bool
Array::unique(size_t pos, const Item &item)
{
	while (pos-- > 0)
		if (items[pos] == item)
			return (false);

	return (true);
}

Array::Array(size_t n)
{
	items = new Item[len = n];
}

Array::Array(const Array &a)
{
	items = new Item[len = a.len];
	memcpy(items, a.items, sizeof(Item) * len);
}

Array &
Array::operator = (const Array &a)
{
	/* do we need to realloc? */
	if (len != a.len) {
		delete [] items;
		items = new Item[len = a.len];
	}

	memcpy(items, a.items, sizeof(Item) * len);
}

Array &
Array::operator = (const Array *a)
{
	/* do we need to realloc? */
	if (len != a->len) {
		delete [] items;
		items = new Item[len = a->len];
	}

	memcpy(items, a->items, sizeof(Item) * len);
}

Item &
Array::operator [] (size_t index) const
{
	assert(index < len);

	return (items[index]);
}

void
Array::reset(Order order)
{
	switch (order) {
	case Array::ASCEND:
		for (size_t i = 0; i < len; i++)
			items[i] = (Item)(i + 1);
		break;

	case Array::DESCEND:
		for (size_t i = 0; i < len; i++)
			items[i] = (Item)(len - i);
		break;

	/* defaults to Array::RANDOM */
	default:
		/* XXX This is quite slow */
		for (size_t i = 0; i < len; i++) {
			Item item;

			/* generate a uniqe random number in [1...len] */
			do {
				item = Item::random(len);
			} while (!unique(i, item));

			items[i] = item;
		}
		break;
	}
}

