/*
 * sortfctns.cpp - sort functions
 *
 * 2003  Benedikt Meurer <bmeurer@unix-ag.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdlib.h>
#include <string.h>

#include "array.h"
#include "util.h"
#include "sortfctns.h"


/* prototypes */
void bubbleSort(Array &);
void bubbleSortBi(Array &);
void selectionSort(Array &);
void insertionSort(Array &);
void shellSort(Array &);
void mergeSort(Array &);
void quickSortLom(Array &);
void quickSortRnd(Array &);
void quickSortHoare(Array &);
void heapSort(Array &);

/**
 * Whenever you add a new sorting algorithm, you need to add an entry to
 * this list.
 */
SortFunctionDescr sortFunctions[] =
{
	/* function       identifier        name */  
	{ bubbleSort,     "bubblesort",     "Bubblesort" },
	{ bubbleSortBi,   "bubblesortbi",   "Bubblesort (bidirectional)" },
	{ selectionSort,  "selectionsort",  "Selectionsort" },
	{ insertionSort,  "insertionsort",  "Insertionsort" },
	{ shellSort,      "shellsort",      "Shellsort" },
	{ mergeSort,      "mergesort",      "Mergesort" },
	{ quickSortLom,   "quicksort",      "Quicksort (Lomuto)" },
	{ quickSortRnd,   "quicksortrnd",   "Quicksort (Randomized)" },
	{ quickSortHoare, "quicksorthoare", "Quicksort (Hoare)" },
	{ heapSort,       "heapsort",       "Heapsort" },
	{ 0, 0, 0 }
};

/**
 * string list of avaiable sort algorithms (default, when no -a is given)
 */
char *
getAvailableAlgorithms(void)
{
	const SortFunctionDescr *p;
	static char *strlist = 0;
	unsigned len = 0;

	if (strlist)
		return (strlist);

	/* calculate required str length */
	for (p = sortFunctions; p->fctn != 0; p++)
		len += strlen(p->identifier) + 1;

	strlist = new char[len + 1];
	strlist[0] = '\0';

	for (p = sortFunctions; p->fctn != 0; p++) {
		strncat(strlist, p->identifier, len + 1);
		strncat(strlist, ",", len + 1);
	}

	/* strip the last ',' */
	strlist[len - 1] = '\0';

	return (strlist);
}


/******************************************************************************
 *									      *
 *                  Sorting algorithms start here.                            *
 *									      *
 *****************************************************************************/

void
bubbleSort(Array &a)
{
	int i, offset = 0;
	bool swapped;

	do {
		swapped = false;

		for (i = 1; i < a.length() - offset; i++) {
			if (a[i - 1] > a[i]) {
				swap(a, i - 1, i);
				swapped = true;
			}
		}

		offset++;
	} while (swapped);
}

/**
 * Bubblesort (bidirectional)
 * array is sorted from beginning to end and versa at same time.
 */
void
bubbleSortBi(Array &a)
{
	int i, st, elements;
	bool ready;

	elements = a.length() - 1;
	ready = false;

	for (st = 0; st < elements && !ready; st++, elements--) {
		ready = true;

		/* left bubblesort */
		for (i = st; i < elements; i++) {
			if (a[i] > a[i + 1]) {
				swap(a, i, i + 1);
				ready = false;
			}
		}

		/* right bubblesort */
		for (i = elements - 1; i >= st; i--) {
			if (a[i] > a[i + 1]) {
				swap(a, i, i + 1);
				ready = false;
			}
		}
	}
}

void
selectionSort(Array &a)
{
	for (int i = 0; i < a.length() - 1; i++) {
		int minIndex = i;

		for (int j = i + 1; j < a.length(); j++)
			if (a[minIndex] > a[j])
				minIndex = j;

		swap(a, i, minIndex);
	}
}

void
insertionSort(Array &a)
{
	int i, j;

	for (i = 1; i < a.length(); i++) {
		Item savedItem = a[i];

		for (j = i - 1; j >= 0 && a[j] > savedItem; j--)
			a[j + 1] = a[j];

		a[j + 1] = savedItem;
	}
}

void
shellSort(Array &a)
{
	int h = 1, i, j;

	while (h < a.length() / 9)
		h = 3 * h + 1;

	for (; h > 0; h /= 3) {
		for (i = h; i < a.length(); i++) {
			Item savedItem = a[i];

			for (j = i; j >= h && a[j - h] > savedItem; j -= h)
				a[j] = a[j - h];

			a[j] = savedItem;
		}
	}
}

static void
merge(Array &a, Array &b, int l, int m, int r)
{
	int i = l;
	int j = m;

	for (int k = l; k < r; k++)
		if ((i >= m) || (j < r && a[i] > a[j])) {
			b[k] = a[j];
			j++;
		}
		else {
			b[k] = a[i];
			i++;
		}

	for (int k = l; k < r; k++)
		a[k] = b[k];
}

static void
mergeSortIntern(Array &a, Array &b, int l, int r)
{
	if (r - l > 1) {
		int m = (l + r) / 2;

		mergeSortIntern(a, b, l, m);
		mergeSortIntern(a, b, m, r);

		merge(a, b, l, m, r);
	}
}

void
mergeSort(Array &a)
{
	Array b(a.length());

	mergeSortIntern(a, b, 0, a.length());
}

/**
 * Partition algorithm as mentioned by Nico Lomuto.
 */
static int
partitionLom(Array &a, int l, int r)
{
	Item pivotItem = a[r];
	int i = l - 1;

	for (int j = l; j < r; j++)
		if (a[j] <= pivotItem) {
			i++;
			swap(a, i, j);
		}

	swap(a, i + 1, r);

	return (i + 1);
}

/**
 * The internal quicksort algorithm (Lomuto)
 */
static void
quickSortInternLom(Array &a, int l, int r)
{
	if (r - l > 0) {
		int pivotIndex = partitionLom(a, l, r);

		quickSortInternLom(a, l, pivotIndex - 1);
		quickSortInternLom(a, pivotIndex + 1, r);
	}
}

/**
 * Quicksort (Lomuto)
 */
void
quickSortLom(Array &a)
{
	quickSortInternLom(a, 0, a.length() - 1);
}

/**
 * The internal quicksort algorithm (Lomuto with randomized median)
 */
void
quickSortInternRnd(Array &a, int l, int r)
{
	if (r - l > 0) {
		/* generate random index in [l...(r-1)] */
		int rndIndex = l + rnd(r - l);

		/* move random median to the end, so we need not to modify
		 * the partitionVah().
		 */
		swap(a, rndIndex, r);

		int pivotIndex = partitionLom(a, l, r);
		quickSortInternRnd(a, l, pivotIndex - 1);
		quickSortInternRnd(a, pivotIndex + 1, r);
	}
}

/**
 * Quicksort (Lomuto with randomized median)
 */
void
quickSortRnd(Array &a)
{
	quickSortInternRnd(a, 0, a.length() - 1);
}

/**
 * Partition as originally introduced by Hoare.
 */
static int
partitionHoare(Array &a, int l, int r)
{
	Item pivotItem = a[l];
	int i = l - 1;
	int j = r + 1;

	for (;;) {
		do {
			j--;
		} while (a[j] > pivotItem);

		do {
			i++;
		} while (a[i] < pivotItem);

		if (i < j)
			swap(a, i, j);
		else
			return (j);
	}
}

/**
 * Internal quicksort algorithm (Hoare)
 */
static void
quickSortInternHoare(Array &a, int l, int r)
{
	if (r - l > 0) {
		int pivotIndex = partitionHoare(a, l, r);

		quickSortInternHoare(a, l, pivotIndex);
		quickSortInternHoare(a, pivotIndex + 1, r);
	}
}

/**
 * Quicksort (Hoare)
 */
void
quickSortHoare(Array &a)
{
	quickSortInternHoare(a, 0, a.length() - 1);
}

/**
 * Create heap for heapsort
 */
static int left(int i)	{ return (2 * i + 1); }
static int right(int i)	{ return (2 * i + 2); }

static void
heapify(Array &a, int i, int size)
{
	int j, l, r;
	int largest;

	for (j = i;;) {
		l = left(j);
		r = right(j);

		if (l < size && a[l] > a[j])
			largest = l;
		else
			largest = j;

		if (r < size && a[r] > a[largest])
			largest = r;

		if (largest != j) {
			swap(a, j, largest);
			largest = j;
		}
		else
			break;
	}
}

/**
 * Heapsort (traditional recursive)
 */
void
heapSort(Array &a)
{
	int size = a.length() - 1;
	int i;

	/* create max heap */
	for (i = a.length() / 2 - 1; i >= 0; i--)
		heapify(a, i, size);

	/* do sorting work */
	for (i = a.length() - 1; i >= 1; i--) {
		swap(a, 0, i);
		heapify(a, 0, size--);
	}
}

